@@ -59,7 +59,7 @@ shutdown = False
 
 def usage():
     print("""%s version %s
-Usage: %s [-346aDeEFhLmqnrvTQxXHq] [-i interval] [-w wait] [-p dst_port] [-P src_port] [-S src_ip]
+Usage: %s [-346aDeEFhLmqnrvTQxXH] [-i interval] [-w wait] [-p dst_port] [-P src_port] [-S src_ip]
        %s [-c count] [-t qtype] [-C class] [-s server] [--ecs client_subnet] hostname
 
   -h, --help        Show this help message
@@ -86,12 +86,13 @@ Usage: %s [-346aDeEFhLmqnrvTQxXHq] [-i interval] [-w wait] [-p dst_port] [-P src
   -C, --class       DNS request record class (default: IN)
   -a, --answer      Display the first matching answer in rdata, if applicable
   -e, --edns        Enable EDNS0 and set its options
-  -E, --ede         Display EDE (Extended DNS Error) messages, when available
+  -E, --ede         (Ignored - EDE messages now always displayed when present)
   -n, --nsid        Enable the NSID bit to retrieve resolver identification (implies EDNS)
+      --cookie      Display EDNS cookies when present
   -D, --dnssec      Enable the DNSSEC desired flag (implies EDNS)
       --ecs         Set EDNS Client Subnet option (format: IP/prefix, e.g., 192.168.1.0/24)
   -F, --flags       Display response flags
-  -x, --expert      Display additional information (implies --ttl, --flags, --ede)
+  -x, --expert      Display additional information (implies --ttl, --flags)
 """ % (__progname__, __version__, __progname__, ' ' * len(__progname__)))
     sys.exit(0)
 
@@ -120,6 +121,7 @@ def print_stderr(s, should_die):
 def validate_server_address(dnsserver, address_family):
     """checks if we have a valid dns server address and resolve if it is a hostname"""
 
+    original_server = dnsserver
     try:
         ipaddress.ip_address(dnsserver)
     except ValueError:  # so it is not a valid IPv4 or IPv6 address, so try to resolve host name
@@ -127,10 +129,11 @@ def validate_server_address(dnsserver, address_family):
             dnsserver = socket.getaddrinfo(dnsserver, port=None, family=address_family)[1][4][0]
         except OSError:
             print_stderr('Error: cannot resolve hostname: %s' % dnsserver, True)
-    return dnsserver
+    return dnsserver, original_server
 
 
 def main():
+    global shutdown
     setup_signal_handler()
 
     if len(sys.argv) == 1:
@@ -146,7 +149,7 @@ def main():
     quiet = False
     verbose = False
     show_flags = False
-    show_ede = False
+    show_cookie = False
     dnsserver = None  # do not try to use system resolver by default
     proto = PROTO_UDP
     dst_port = getDefaultPort(proto)
@@ -169,7 +172,7 @@ def main():
                                    ["help", "count=", "server=", "quiet", "type=", "wait=", "interval=", "verbose",
                                     "port=", "srcip=", "tcp", "ipv4", "ipv6", "cache-miss", "srcport=", "edns",
                                     "dnssec", "flags", "norecurse", "tls", "doh", "nsid", "ede", "class=", "ttl",
-                                    "expert", "answer", "quic", "http3", "ecs="])
+                                    "expert", "answer", "quic", "http3", "ecs=", "cookie"])
     except getopt.GetoptError as err:
         # print help information and exit:
         print_stderr(err, False)  # will print something like "option -a not recognized"
@@ -208,7 +211,6 @@ def main():
 
         elif o in ("-x", "--expert"):
             show_flags = True
-            show_ede = True
             show_ttl = True
 
         elif o in ("-m", "--cache-miss"):
@@ -278,7 +280,11 @@ def main():
             show_flags = True
 
         elif o in ("-E", "--ede"):
-            show_ede = True
+            # Kept for compatibility - EDE now always displayed when present
+            pass
+
+        elif o == "--cookie":
+            show_cookie = True
 
         elif o in ("-p", "--port"):
             dst_port = int(a)
@@ -304,7 +310,7 @@ def main():
     if dnsserver is None:
         dnsserver = dns.resolver.get_default_resolver().nameservers[0]
 
-    dnsserver = validate_server_address(dnsserver, af)
+    dnsserver_ip, dnsserver_hostname = validate_server_address(dnsserver, af)
 
     response_time = []
     i = 0
@@ -313,8 +319,10 @@ def main():
     if not valid_rdatatype(rdatatype):
         print_stderr('Error: Invalid record type: %s ' % rdatatype, True)
 
+    # Display the hostname if it differs from the resolved IP, otherwise just the IP
+    server_display = dnsserver_hostname if dnsserver_hostname != dnsserver_ip else dnsserver_ip
     print("%s DNS: %s:%d, hostname: %s, proto: %s, class: %s, type: %s, flags: [%s]" %
-          (__progname__, dnsserver, dst_port, qname, proto_to_text(proto), dns.rdataclass.to_text(rdata_class),
+          (__progname__, server_display, dst_port, qname, proto_to_text(proto), dns.rdataclass.to_text(rdata_class),
            rdatatype, dns.flags.to_text(request_flags)), flush=True)
 
     while not shutdown:
@@ -339,6 +347,11 @@ def main():
                     edns_options.append(ecs_option)
                 except Exception as e:
                     print_stderr("Error: Invalid ECS format '%s': %s" % (client_subnet, e), True)
+            if show_cookie:
+                # Send a client cookie (8 random bytes as per RFC 7873)
+                import os
+                client_cookie = os.urandom(8)
+                edns_options.append(dns.edns.GenericOption(10, client_cookie))  # COOKIE = 10
 
             query = dns.message.make_query(fqdn, rdatatype, rdata_class, flags=request_flags,
                                            use_edns=True, want_dnssec=want_dnssec, payload=1232,
@@ -350,22 +363,30 @@ def main():
         try:
             stime = time.perf_counter()
             if proto is PROTO_UDP:
-                answers = dns.query.udp(query, dnsserver, timeout=timeout, port=dst_port,
+                answers = dns.query.udp(query, dnsserver_ip, timeout=timeout, port=dst_port,
                                         source=src_ip, source_port=src_port, ignore_unexpected=True)
             elif proto is PROTO_TCP:
-                answers = dns.query.tcp(query, dnsserver, timeout=timeout, port=dst_port,
+                answers = dns.query.tcp(query, dnsserver_ip, timeout=timeout, port=dst_port,
                                         source=src_ip, source_port=src_port)
             elif proto is PROTO_TLS:
                 if hasattr(dns.query, 'tls'):
-                    answers = dns.query.tls(query, dnsserver, timeout=timeout, port=dst_port,
-                                            source=src_ip, source_port=src_port)
+                    # Use resolved IP for connection, but provide hostname for SNI/certificate validation
+                    server_hostname = dnsserver_hostname if dnsserver_hostname != dnsserver_ip else None
+                    answers = dns.query.tls(query, dnsserver_ip, timeout=timeout, port=dst_port,
+                                            source=src_ip, source_port=src_port,
+                                            server_hostname=server_hostname)
                 else:
                     unsupported_feature("DNS-over-TLS")
 
             elif proto is PROTO_HTTPS:
                 if hasattr(dns.query, 'https'):
                     try:
-                        answers = dns.query.https(query, dnsserver, timeout=timeout, port=dst_port,
+                        # For hostnames, construct full URL; for IPs, use IP directly
+                        if dnsserver_hostname != dnsserver_ip:
+                            https_server = f"https://{dnsserver_hostname}/dns-query"
+                        else:
+                            https_server = dnsserver_ip
+                        answers = dns.query.https(query, https_server, timeout=timeout, port=dst_port,
                                                   source=src_ip, source_port=src_port)
                     except httpx.ConnectError:
                         print_stderr(f"The server did not respond to DoH on port {dst_port}", should_die=True)
@@ -375,7 +396,12 @@ def main():
             elif proto is PROTO_HTTP3:
                 if hasattr(dns.query, 'quic'):
                     try:
-                        answers = dns.query.https(query, dnsserver, timeout=timeout, port=dst_port,
+                        # For hostnames, construct full URL; for IPs, use IP directly
+                        if dnsserver_hostname != dnsserver_ip:
+                            https_server = f"https://{dnsserver_hostname}/dns-query"
+                        else:
+                            https_server = dnsserver_ip
+                        answers = dns.query.https(query, https_server, timeout=timeout, port=dst_port,
                                                   source=src_ip, source_port=src_port,
                                                   http_version=dns.query.HTTPVersion.H3)
                     except ConnectionRefusedError:
@@ -387,8 +413,11 @@ def main():
             elif proto is PROTO_QUIC:
                 if hasattr(dns.query, 'quic'):
                     try:
-                        answers = dns.query.quic(query, dnsserver, timeout=timeout, port=dst_port,
-                                                 source=src_ip, source_port=src_port)
+                        # Use resolved IP for connection, but provide hostname for SNI/certificate validation
+                        server_hostname = dnsserver_hostname if dnsserver_hostname != dnsserver_ip else None
+                        answers = dns.query.quic(query, dnsserver_ip, timeout=timeout, port=dst_port,
+                                                 source=src_ip, source_port=src_port,
+                                                 server_hostname=server_hostname)
                     except dns.exception.Timeout:
                         print_stderr(f"The server did not respond to DoQ on port {dst_port}", should_die=True)
                     except ConnectionRefusedError:
@@ -422,6 +451,10 @@ def main():
             if not quiet:
                 print_stderr("Invalid Response", False)
                 continue
+        except KeyboardInterrupt:
+            # Handle Ctrl+C during DNS query
+            shutdown = True
+            break
         else:
             # convert time to milliseconds, considering that
             # time property is returned differently by query.https
@@ -456,7 +489,8 @@ def main():
                             if nsid_val:
                                 edns_parts.append("NSID:%s" % nsid_val.decode("utf-8"))
 
-                if show_ede and answers.options:
+                # Always show EDE when present (server-initiated error responses)
+                if answers.options:
                     for ans_opt in answers.options:
                         if ans_opt.otype == dns.edns.EDE:
                             if ans_opt.text:
@@ -471,10 +505,16 @@ def main():
                     for ans_opt in answers.options:
                         if ans_opt.otype == dns.edns.OptionType.ECS:
                             if ans_opt.address:
-                                edns_parts.append("ECS:%s/%d" % (ans_opt.address, ans_opt.scopelen or ans_opt.srclen))
+                                edns_parts.append("ECS:%s/%d/%d" % (ans_opt.address, ans_opt.srclen, ans_opt.scopelen))
                             else:
                                 edns_parts.append("ECS:auto")
 
+                # Always show cookies when present (echoed back from server)
+                if answers.options:
+                    for ans_opt in answers.options:
+                        if ans_opt.otype == 10:  # COOKIE
+                            edns_parts.append("COOKIE:%d" % len(ans_opt.data))
+
                 if edns_parts:
                     extras += " [%s]" % ", ".join(edns_parts)
 
@@ -485,7 +525,7 @@ def main():
                             break
 
                 print("%-3d bytes from %s: seq=%-3d time=%-7.3f ms %s" % (
-                    len(answers.to_wire()), dnsserver, i, elapsed, extras), flush=True)
+                    len(answers.to_wire()), server_display, i, elapsed, extras), flush=True)
 
             if verbose:
                 print(answers.to_text(), flush=True)
@@ -539,9 +579,18 @@ def main():
 
                         print("%s (%d): %s" % (option_name, ans_opt.otype, option_details), flush=True)
 
+            # Check for shutdown signal after verbose output processing
+            if shutdown:
+                break
+
             time_to_next = (stime + interval) - etime
             if time_to_next > 0:
-                time.sleep(time_to_next)
+                # Interruptible sleep - check for shutdown signal every 0.1 seconds
+                sleep_start = time.time()
+                while time.time() - sleep_start < time_to_next:
+                    if shutdown:
+                        break
+                    time.sleep(min(0.1, time_to_next - (time.time() - sleep_start)))
 
     r_sent = i
     r_received = len(response_time)
@@ -561,7 +610,7 @@ def main():
         r_avg = 0
         r_stddev = 0
 
-    print('\n--- %s dnsping statistics ---' % dnsserver, flush=True)
+    print('\n--- %s dnsping statistics ---' % server_display, flush=True)
     print('%d requests transmitted, %d responses received, %.0f%% lost' % (r_sent, r_received, r_lost_percent),
           flush=True)
     print('min=%.3f ms, avg=%.3f ms, max=%.3f ms, stddev=%.3f ms' % (r_min, r_avg, r_max, r_stddev), flush=True)
